import sys
#from mext.addons import *
from weakref import ref #, proxy
from types import MethodType, FunctionType, ClassType


__all__ = [
    #'AddOn', 'Registry',
    'BindingProperty', 'BindingDescriptor', 'GeneratorWrapper',
    'class_assign', 'init_args', 'inject', 'mix_into',
    'alias', 'filter_alias', #'tuple_property',
    'set_once', 'property_once',
    'classy', 'on_class_ready', 'is_in_class_declaration',
]



def super_next(cls, attr):
    for c in cls.__mro__:
        if attr in c.__dict__:
            yield getattr(c, attr).im_func

class classy_class(type):
    """Metaclass that delegates selected operations back to the class"""

    def __new__(meta, name, bases, cdict):
        cls = super(classy_class, meta).__new__(meta, name, bases, cdict)
        supr = super_next(cls, '__class_new__').next
        return supr()(meta, name, bases, cdict, supr)

    def __init__(cls, name, bases, cdict):
        supr = super_next(cls, '__class_init__').next
        return supr()(cls, name, bases, cdict, supr)

    def __call__(cls, *args, **kw):
        return cls.__class_call__.im_func(cls, *args, **kw)


class classy(object):
    """Base class for classes that want to be their own metaclass"""
    __metaclass__ = classy_class
    __slots__ = ()

    def __class_new__(meta, name, bases, cdict, supr):
        return type.__new__(meta, name, bases, cdict)

    def __class_init__(cls, name, bases, cdict, supr):
        return type.__init__(cls, name, bases, cdict)

    def __class_call__(cls, *args, **kw):
        return type.__call__(cls, *args, **kw)
    __class_call__ = classmethod(__class_call__)












def bind(rule, ob):
    if hasattr(rule, '__get__'):
        return rule.__get__(ob, type(ob))
    return rule

class alias(object):
    __slots__ = 'path'
    def __init__(self, path):
        if isinstance(path, str):
            self.path = tuple(path.split('.'))
        elif isinstance(path, (list, tuple)):
            self.path = tuple(path)
        else:
            raise TypeError(path)

    def __repr__(self):
        return '%s.%s(%r)' % (self.__class__.__module__, self.__class__.__name__, '.'.join(self.path))

    @staticmethod
    def get(ob, path):
        for name in path:
            ob = getattr(ob, name)
        return ob

    @classmethod
    def set(cls, ob, path, value):
        ob = cls.get(ob, path[:-1])
        setattr(ob, path[-1], value)

    def __get__(self, ob, type):
        if ob is None:
            return self
        return self.get(ob, self.path)

    def __set__(self, ob, value):
        self.set(ob, self.path, value)


class filter_alias(alias):
    def __init__(self, name, decode, encode):
        alias.__init__(self, name)
        self.decode = decode
        self.encode = encode

    def __get__(self, ob, typ):
        if ob is None:
            return self
        val = alias.__get__(self, ob, typ)
        decode = bind(self.decode, ob)
        return decode(val)

    def __set__(self, ob, val):
        encode = bind(self.encode, ob)
        val = encode(val)
        alias.__set__(self, ob, val)


## def tuple_property(name1, name2):
##     def get(self):
##         return getattr(self, name1), getattr(self, name2)
##     def set(self, value):
##         assert isinstance(value, tuple) and len(value) == 2
##         setattr(self, name1, value[0])
##         setattr(self, name2, value[1])
##     return property(get, set)
##


def get_name(ob, default=TypeError):
    if isinstance(ob, (type, ClassType)):
        return ob.__name__
    elif isinstance(ob, FunctionType):
        return ob.func_name
    elif isinstance(ob, MethodType):
        return get_name(ob.im_func, default)
    elif isinstance(ob, (classmethod, staticmethod)):
        return get_name(ob.__get__(None, object), default)
    elif isinstance(ob, property):
        return get_name(ob.fget, default)
    elif default is not TypeError:
        return default
    else:
        raise TypeError("Can't get name of %r" % ob)

def inject(parent, setname=None, save=False):
    def deco(ob):
        if setname is None:
            name = get_name(ob)
        else:
            name = setname
        if save:
            if isinstance(save, str):
                prev_name = save
            else:
                prev_name = '_%s_%s' % (parent.__name__, name)
            setattr(parent, prev_name, getattr(parent, name))
        setattr(parent, name, ob)
        return ob
    return deco


def init_args(*argnames):
    @on_class_ready
    def set_init(cls):
        def init(self, *args, **kw):
            for name, val in zip(argnames, args):
                if name in kw:
                    raise ValueError('Duplicate argument: %s' % name)
                kw[name] = val
            super(cls, self).__init__(**kw)
        cls.__init__ = init



def mix_into(cls, mixin=None):
    if mixin is not None:
        cls.__bases__ = (mixin,) + cls.__bases__
    else:
        on_class_ready(lambda mixin: mix_into(cls, mixin))




def class_assign(key, value, depth=2):
    frame = enclosing_frame(None, depth+1)
    kind, module, caller_locals, caller_globals = frameinfo(frame)
    if kind != "class":
        #print caller_locals
        raise SyntaxError("Class decorators may only be used inside a class statement; kind=%s" % kind)
    assert key not in caller_locals
    caller_locals[key] = value





class property_once(object):
    __slots__ = 'make'
    def __init__(self, make):
        self.make = make

    def __get__(self, ob, type):
        if type is None:
            return self
        try:
            return ob.__dict__[self]
        except KeyError:
            return ob.__dict__.setdefault(self, self.make.__get__(ob, type)())

_not_set = object()

class set_once(property_once):
    __slots__ = 'make'
    def __init__(self, default=_not_set, make=None):
        #assert default is _not_set or make is None
        #self.default = default
        if make is None:
            make = lambda self: default
        else:
            assert default is _not_set
        property_once.__init__(self, make)

    def __get__(self, ob, type):
        value = property_once.__get__(self, ob, type)
        if value is _not_set:
            raise AttributeError("Attribute not set yet")
        return value

    def __set__(self, ob, value):
        if self in ob.__dict__:
            if ob.__dict__[self] != value:
                raise AttributeError("This attribute is already set")
        else:
            ob.__dict__.setdefault(self, value)



class GeneratorWrapper(object):
    def __init__(self, iterator):
        self.iterator = iterator
        self.next = self._wrap_method(iterator.next)
        if hasattr(iterator, 'send'):
            self.send = self._wrap_method(iterator.send)
        if hasattr(iterator, 'throw'):
            self.throw = self._wrap_method(iterator.throw)

    def _wrap_method(self, method):
        return lambda *args: self.call_wrap(method, *args)

    def call_wrap(self, method, *args):
        raise NotImplementedError






def is_in_class_declaration(depth):
    try:
        return '__module__' in sys._getframe(depth+1).f_locals
    except ValueError:
        return False






def on_class_ready(cb, depth=1):
    frame = sys._getframe(depth+1)
    if '__module__' not in frame.f_locals:
        raise SyntaxError("Not in class declaration")
    if '__metaclass__' in frame.f_globals:
        raise SyntaxError("Module-wide metaclasses are not supported")

    meta_prev = frame.f_locals.get('__metaclass__')
    callbacks = getattr(meta_prev, '_on_class_ready_callbacks', None)
    if callbacks is not None:
        callbacks.append(cb)
        return

    def meta_cb(name, bases, cdict):
        cdict.pop('__metaclass__', None)
        meta = metaclass_for_bases(bases, meta_prev)
        r = meta(name, bases, cdict)
        for cb in callbacks:
            cb(r)
        return r

    callbacks = meta_cb._on_class_ready_callbacks = [cb]
    frame.f_locals['__metaclass__'] = meta_cb


def metaclass_for_bases(bases, explicit_mc=None):
    meta = map(type, bases)
    if explicit_mc is not None:
        meta.append(explicit_mc)
    if ClassType in meta:
        raise TypeError("Classic classes are not supported: %s" % [b for b in bases if isinstance(b, ClassType)])
    r = None
    for m in meta:
        if r is None or issubclass(m, r):
            r = m
        elif not issubclass(r, m):
            raise TypeError("Incompatible metatypes", bases)
    return r or ClassType














#@@
from mext.addons import AddOn

class BindingDescriptor(object):
    def __init__(self, attr_cls, **kw):
        self.attr_cls = attr_cls
        self._extra_kw_repr = ', '.join(map('%s=%r'.__mod__, kw.items()))
        for key, value in kw.items():
            setattr(self, key, value)

    def __repr__(self):
        return self.__class__.__name__ + '(%s)' % self._extra_kw_repr

    def __get__(self, obj, type):
        if obj is None:
            return self
        return self.attr_cls(obj, self).get()

    def __set__(self, obj, value):
        self.attr_cls(obj, self).set(value)

    def __delete__(self):
        self.attr_cls(obj, self).delete()

    def exists_for(self, ob):
        return self.attr_cls.exists_for(ob, self)

    def get_for(self, ob):
        return self.attr_cls(ob, self)


class BindingProperty(AddOn):
    def __init__(self, ob, attr):
        #self.ob = proxy(ob)
        self._ref = ref(ob)
        self.attr = attr
        self.init()

    ob = property(lambda self: self._ref())

    def init(self):
        pass

    def get(self):
        raise NotImplementedError

    def set(self, value):
        raise NotImplementedError

    def delete(self, value):
        raise NotImplementedError

    def __repr__(self):
        return self.__class__.__name__ + repr((self.ob, self.attr))

    @classmethod
    def make_attr(cls, **kw):
        return BindingDescriptor(cls, **kw)


